# -*- coding: utf-8 -*-
"""
Created on Sat Sep 10 21:00:28 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest
from pathlib import Path
from collections import OrderedDict

import numpy as np

from mth5.io.zen import Z3D

from mt_metadata.timeseries.filters import (
    ChannelResponseFilter,
    FrequencyResponseTableFilter,
    CoefficientFilter,
)

# =============================================================================


@unittest.skipIf("peacock" not in str(Path(__file__).as_posix()), "local files")
class TestZ3DEY(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.fn = Path(
            r"c:\Users\jpeacock\OneDrive - DOI\mt\example_z3d_data\bm100_20220517_131017_256_EY.Z3D"
        )
        self.z3d = Z3D(self.fn)
        self.z3d.read_z3d()

        self.maxDiff = None

    def test_fn(self):
        self.assertEqual(self.fn, self.z3d.fn)

    def test_file_size(self):
        self.assertEqual(self.z3d.file_size, 10759100)

    def test_n_samples(self):
        self.assertEqual(self.z3d.n_samples, 2530304)

    def test_station(self):
        self.assertEqual(self.z3d.station, "100")

    def test_dipole_length(self):
        self.assertEqual(self.z3d.dipole_length, 56.0)

    def test_azimuth(self):
        self.assertEqual(self.z3d.azimuth, 0)

    def test_component(self):
        self.assertEqual(self.z3d.component, "ey")

    def test_latitude(self):
        self.assertAlmostEqual(self.z3d.latitude, 40.49757833327694)

    def test_longitude(self):
        self.assertAlmostEqual(self.z3d.longitude, -116.8211900230401)

    def test_elevation(self):
        self.assertAlmostEqual(self.z3d.elevation, 1456.3)

    def test_gps_stamps(self):
        self.assertEqual(9885, self.z3d.gps_stamps.size)

    def test_sample_rate(self):
        self.assertEqual(self.z3d.sample_rate, 256)

    def test_start(self):
        self.assertEqual(self.z3d.start, "2022-05-17T13:09:58+00:00")

    def test_end(self):
        self.assertEqual(self.z3d.end, "2022-05-17T15:54:42+00:00")

    def test_zen_schedule(self):
        self.assertEqual(self.z3d.zen_schedule, "2022-05-17T13:09:58+00:00")

    def test_coil_number(self):
        self.assertEqual(self.z3d.coil_number, None)

    def test_channel_number(self):
        self.assertEqual(self.z3d.channel_number, 5)

    def test_gps_stamps_seconds(self):
        self.assertEqual(
            self.z3d.gps_stamps.size - 1, self.z3d.end - self.z3d.start
        )

    def test_get_gps_stamp_type(self):
        self.assertEqual(
            np.dtype(
                [
                    ("flag0", "<i4"),
                    ("flag1", "<i4"),
                    ("time", "<i4"),
                    ("lat", "<f8"),
                    ("lon", "<f8"),
                    ("gps_sens", "<i4"),
                    ("num_sat", "<i4"),
                    ("temperature", "<f4"),
                    ("voltage", "<f4"),
                    ("num_fpga", "<i4"),
                    ("num_adc", "<i4"),
                    ("pps_count", "<i4"),
                    ("dac_tune", "<i4"),
                    ("block_len", "<i4"),
                ]
            ),
            self.z3d._gps_dtype,
        )

    def test_gps_stamp_length(self):
        self.assertEqual(self.z3d._gps_stamp_length, 64)

    def test_gps_bytes(self):
        self.assertEqual(self.z3d._gps_bytes, 16)

    def test_gps_flag_0(self):
        self.assertEqual(self.z3d._gps_flag_0, 2147483647)

    def test_gps_flag_1(self):
        self.assertEqual(self.z3d._gps_flag_1, -2147483648)

    def test_block_len(self):
        self.assertEqual(self.z3d._block_len, 65536)

    def test_gps_flag(self):
        self.assertEqual(self.z3d.gps_flag, b"\xff\xff\xff\x7f\x00\x00\x00\x80")

    def test_get_gps_time(self):
        self.assertTupleEqual(
            self.z3d.get_gps_time(220216, 2210), (215.056, 2210.0)
        )

    def test_get_utc_date_time(self):
        self.assertEqual(
            self.z3d.get_UTC_date_time(2210, 220216),
            "2022-05-17T13:09:58+00:00",
        )

    def test_channel_metadata(self):
        ey = OrderedDict(
            [
                ("ac.end", 3.1080129002282815e-05),
                ("ac.start", 3.4870685796509496e-05),
                ("channel_number", 5),
                ("component", "ey"),
                ("data_quality.rating.value", 0),
                ("dc.end", 0.019371436521409924),
                ("dc.start", 0.019130984313785026),
                ("dipole_length", 56.0),
                ("filter.applied", [False, False]),
                (
                    "filter.name",
                    ["zen_counts2mv", "dipole_56.00m"],
                ),
                ("measurement_azimuth", 0.0),
                ("measurement_tilt", 0.0),
                ("negative.elevation", 0.0),
                ("negative.id", None),
                ("negative.latitude", 0.0),
                ("negative.longitude", 0.0),
                ("negative.manufacturer", None),
                ("negative.type", None),
                ("positive.elevation", 0.0),
                ("positive.id", None),
                ("positive.latitude", 0.0),
                ("positive.longitude", 0.0),
                ("positive.manufacturer", None),
                ("positive.type", None),
                ("sample_rate", 256.0),
                ("time_period.end", "2022-05-17T15:54:42+00:00"),
                ("time_period.start", "2022-05-17T13:09:58+00:00"),
                ("type", "electric"),
                ("units", "digital counts"),
            ]
        )

        self.assertDictEqual(self.z3d.channel_metadata.to_dict(single=True), ey)

    def test_run_metadata(self):
        rm = OrderedDict(
            [
                ("acquired_by.author", ""),
                ("channels_recorded_auxiliary", []),
                ("channels_recorded_electric", []),
                ("channels_recorded_magnetic", []),
                ("data_logger.firmware.author", None),
                ("data_logger.firmware.name", None),
                ("data_logger.firmware.version", "4147.0"),
                ("data_logger.id", "ZEN024"),
                ("data_logger.manufacturer", "Zonge International"),
                ("data_logger.model", "ZEN"),
                ("data_logger.timing_system.drift", 0.0),
                ("data_logger.timing_system.type", "GPS"),
                ("data_logger.timing_system.uncertainty", 0.0),
                ("data_logger.type", None),
                ("data_type", "MTBB"),
                ("id", None),
                ("sample_rate", 256.0),
                ("time_period.end", "2022-05-17T15:54:42+00:00"),
                ("time_period.start", "2022-05-17T13:09:58+00:00"),
            ]
        )

        self.assertDictContainsSubset(
            self.z3d.run_metadata.to_dict(single=True), rm
        )

    def test_station_metadata(self):
        sm = OrderedDict(
            [
                ("acquired_by.name", ""),
                ("channels_recorded", []),
                ("data_type", "BBMT"),
                ("fdsn.id", "100"),
                ("geographic_name", None),
                ("id", "100"),
                ("location.declination.model", "WMM"),
                ("location.declination.value", 0.0),
                ("location.elevation", 1456.3),
                ("location.latitude", 40.49757833327694),
                ("location.longitude", -116.8211900230401),
                ("orientation.method", None),
                ("orientation.reference_frame", "geographic"),
                ("provenance.creation_time", "1980-01-01T00:00:00+00:00"),
                ("provenance.software.author", None),
                ("provenance.software.name", None),
                ("provenance.software.version", None),
                ("provenance.submitter.email", None),
                ("provenance.submitter.organization", None),
                ("release_license", "CC0-1.0"),
                ("run_list", []),
                ("time_period.end", "2022-05-17T15:54:42+00:00"),
                ("time_period.start", "2022-05-17T13:09:58+00:00"),
            ]
        )

        self.assertDictEqual(self.z3d.station_metadata.to_dict(single=True), sm)

    def test_filters(self):
        zr = FrequencyResponseTableFilter(
            **OrderedDict(
                [
                    ("amplitudes", np.array([1.00153])),
                    ("calibration_date", "1980-01-01"),
                    ("comments", "data logger response read from z3d file"),
                    ("frequencies", np.array([2.0])),
                    ("gain", 1.0),
                    ("instrument_type", None),
                    ("name", "zen024_256_response"),
                    ("phases", np.array([-1.5333299999999999])),
                    ("type", "frequency response table"),
                    ("units_in", "mV"),
                    ("units_out", "mV"),
                ]
            )
        )

        df = CoefficientFilter(
            **OrderedDict(
                [
                    ("calibration_date", "1980-01-01"),
                    ("comments", "convert to electric field"),
                    ("gain", 0.056),
                    ("name", "dipole_56.00m"),
                    ("type", "coefficient"),
                    ("units_in", "mV"),
                    ("units_out", "mV/km"),
                ]
            )
        )

        cf = CoefficientFilter(
            **OrderedDict(
                [
                    ("calibration_date", "1980-01-01"),
                    ("comments", "digital counts to millivolts"),
                    ("gain", 1048576000.000055),
                    ("name", "zen_counts2mv"),
                    ("type", "coefficient"),
                    ("units_in", "count"),
                    ("units_out", "mV"),
                ]
            )
        )

        with self.subTest("test zen response"):
            self.assertEqual(None, self.z3d.zen_response)
            # self.assertDictEqual(
            #     self.z3d.zen_response.to_dict(single=True),
            #     zr.to_dict(single=True),
            # )
        with self.subTest("test_dipole_filter"):

            self.assertDictEqual(
                self.z3d.dipole_filter.to_dict(single=True),
                df.to_dict(single=True),
            )

        with self.subTest("test_conversion_filter"):

            self.assertEqual(
                self.z3d.counts2mv_filter.to_dict(single=True),
                cf.to_dict(single=True),
            )

        with self.subTest("channel_response"):

            cr = ChannelResponseFilter(filters_list=[cf, df])
            self.assertListEqual(
                cr.filters_list, self.z3d.channel_response.filters_list
            )


@unittest.skipIf("peacock" not in str(Path(__file__).as_posix()), "local files")
class TestZ3DHY(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.fn = Path(
            r"c:\Users\jpeacock\OneDrive - DOI\mt\example_z3d_data\bm100_20220517_131017_256_HY.Z3D"
        )
        self.z3d = Z3D(self.fn)
        self.z3d.read_z3d()

        self.maxDiff = None

        self.cr = FrequencyResponseTableFilter(
            **OrderedDict(
                [
                    (
                        "amplitudes",
                        np.array(
                            [
                                0.27649,
                                0.39498,
                                0.59248,
                                0.75047,
                                1.14545,
                                1.54044,
                                2.3304,
                                3.08087,
                                4.62131,
                                6.16174,
                                9.25667,
                                12.3343,
                                18.2536,
                                24.0094,
                                35.0682,
                                44.3682,
                                59.9492,
                                70.4913,
                                83.2834,
                                89.5798,
                                95.028,
                                97.1974,
                                98.8414,
                                99.5336,
                                99.975,
                                100.138,
                                100.246,
                                100.26,
                                100.332,
                                100.33,
                                100.37,
                                100.355,
                                100.414,
                                100.275,
                                100.427,
                                100.382,
                                100.307,
                                100.353,
                                99.8983,
                                99.0787,
                                90.2819,
                                68.2665,
                                27.3347,
                                11.8179,
                                3.70505,
                                1.62308,
                                0.5053,
                                0.24206,
                            ]
                        ),
                    ),
                    ("calibration_date", "1980-01-01"),
                    ("comments", "induction coil response read from z3d file"),
                    (
                        "frequencies",
                        np.array(
                            [
                                7.32422e-04,
                                9.76563e-04,
                                1.46484e-03,
                                1.95313e-03,
                                2.92969e-03,
                                3.90625e-03,
                                5.85938e-03,
                                7.81250e-03,
                                1.17188e-02,
                                1.56250e-02,
                                2.34375e-02,
                                3.12500e-02,
                                4.68750e-02,
                                6.25000e-02,
                                9.37500e-02,
                                1.25000e-01,
                                1.87500e-01,
                                2.50000e-01,
                                3.75000e-01,
                                5.00000e-01,
                                7.50000e-01,
                                1.00000e00,
                                1.50000e00,
                                2.00000e00,
                                3.00000e00,
                                4.00000e00,
                                6.00000e00,
                                8.00000e00,
                                1.20000e01,
                                1.60000e01,
                                2.40000e01,
                                3.20000e01,
                                4.80000e01,
                                6.40000e01,
                                9.60000e01,
                                1.28000e02,
                                1.92000e02,
                                2.56000e02,
                                3.84000e02,
                                5.12000e02,
                                7.68000e02,
                                1.02400e03,
                                1.53600e03,
                                2.04800e03,
                                3.07200e03,
                                4.09600e03,
                                6.14400e03,
                                8.19200e03,
                            ]
                        ),
                    ),
                    ("gain", 1.0),
                    ("instrument_type", None),
                    ("name", "ant4_2324_response"),
                    (
                        "phases",
                        np.array(
                            [
                                1.5702e00,
                                1.5628e00,
                                1.5545e00,
                                1.5496e00,
                                1.5409e00,
                                1.5348e00,
                                1.5262e00,
                                1.5205e00,
                                1.5121e00,
                                1.5062e00,
                                1.4792e00,
                                1.4470e00,
                                1.3855e00,
                                1.3300e00,
                                1.2169e00,
                                1.1118e00,
                                9.2770e-01,
                                7.9150e-01,
                                5.9040e-01,
                                4.6760e-01,
                                3.2250e-01,
                                2.4570e-01,
                                1.6350e-01,
                                1.2190e-01,
                                7.8400e-02,
                                5.5100e-02,
                                3.0300e-02,
                                1.5700e-02,
                                -2.9000e-03,
                                -1.6100e-02,
                                -3.7800e-02,
                                -5.6100e-02,
                                -9.1000e-02,
                                -1.2490e-01,
                                -1.9050e-01,
                                -2.5650e-01,
                                -3.8930e-01,
                                -5.2140e-01,
                                -7.9710e-01,
                                -1.0885e00,
                                -1.7401e00,
                                -2.4110e00,
                                2.9802e00,
                                2.5876e00,
                                2.1997e00,
                                1.9765e00,
                                1.7768e00,
                                1.5969e00,
                            ]
                        ),
                    ),
                    ("type", "frequency response table"),
                    ("units_in", "mV"),
                    ("units_out", "nT"),
                ]
            )
        )

        self.zr = FrequencyResponseTableFilter(
            **OrderedDict(
                [
                    ("amplitudes", np.array([0.999878])),
                    ("calibration_date", "1980-01-01"),
                    ("comments", "data logger response read from z3d file"),
                    ("frequencies", np.array([2.0])),
                    ("gain", 1.0),
                    ("instrument_type", None),
                    ("name", "zen024_256_response"),
                    ("phases", np.array([-1.53334])),
                    ("type", "frequency response table"),
                    ("units_in", "mV"),
                    ("units_out", "mV"),
                ]
            )
        )

        self.cf = CoefficientFilter(
            **OrderedDict(
                [
                    ("calibration_date", "1980-01-01"),
                    ("comments", "digital counts to millivolts"),
                    ("gain", 1048576000.000055),
                    ("name", "zen_counts2mv"),
                    ("type", "coefficient"),
                    ("units_in", "count"),
                    ("units_out", "mV"),
                ]
            )
        )

    def test_channel_metadata(self):
        ey = OrderedDict(
            [
                ("channel_number", 2),
                ("component", "hy"),
                ("data_quality.rating.value", 0),
                ("filter.applied", [False, False]),
                (
                    "filter.name",
                    [
                        "zen_counts2mv",
                        "ant4_2324_response",
                    ],
                ),
                ("h_field_max.end", 0.02879215431213228),
                ("h_field_max.start", 0.03145987892150714),
                ("h_field_min.end", 0.02772834396362159),
                ("h_field_min.start", 0.02886334419250337),
                ("location.elevation", 0.0),
                ("location.latitude", 0.0),
                ("location.longitude", 0.0),
                ("measurement_azimuth", 90.0),
                ("measurement_tilt", 0.0),
                ("sample_rate", 256.0),
                ("sensor.id", ["2324"]),
                ("sensor.manufacturer", ["Geotell"]),
                ("sensor.model", ["ANT-4"]),
                ("sensor.type", ["induction coil"]),
                ("time_period.end", "2022-05-17T15:54:42+00:00"),
                ("time_period.start", "2022-05-17T13:09:58+00:00"),
                ("type", "magnetic"),
                ("units", "digital counts"),
            ]
        )

        self.assertDictEqual(self.z3d.channel_metadata.to_dict(single=True), ey)

    def test_run_metadata(self):
        rm = OrderedDict(
            [
                ("acquired_by.author", ""),
                ("channels_recorded_auxiliary", []),
                ("channels_recorded_electric", []),
                ("channels_recorded_magnetic", []),
                ("data_logger.firmware.author", None),
                ("data_logger.firmware.name", None),
                ("data_logger.firmware.version", "4147.0"),
                ("data_logger.id", "ZEN024"),
                ("data_logger.manufacturer", "Zonge International"),
                ("data_logger.model", "ZEN"),
                ("data_logger.timing_system.drift", 0.0),
                ("data_logger.timing_system.type", "GPS"),
                ("data_logger.timing_system.uncertainty", 0.0),
                ("data_logger.type", None),
                ("data_type", "MTBB"),
                ("id", None),
                ("sample_rate", 256.0),
                ("time_period.end", "2022-05-17T15:54:42+00:00"),
                ("time_period.start", "2022-05-17T13:09:58+00:00"),
            ]
        )

        self.assertDictContainsSubset(
            self.z3d.run_metadata.to_dict(single=True), rm
        )

    def test_station_metadata(self):
        sm = OrderedDict(
            [
                ("acquired_by.name", ""),
                ("channels_recorded", []),
                ("data_type", "BBMT"),
                ("fdsn.id", "100"),
                ("geographic_name", None),
                ("id", "100"),
                ("location.declination.model", "WMM"),
                ("location.declination.value", 0.0),
                ("location.elevation", 1456.3),
                ("location.latitude", 40.49757833327694),
                ("location.longitude", -116.8211900230401),
                ("orientation.method", None),
                ("orientation.reference_frame", "geographic"),
                ("provenance.creation_time", "1980-01-01T00:00:00+00:00"),
                ("provenance.software.author", None),
                ("provenance.software.name", None),
                ("provenance.software.version", None),
                ("provenance.submitter.email", None),
                ("provenance.submitter.organization", None),
                ("release_license", "CC0-1.0"),
                ("run_list", []),
                ("time_period.end", "2022-05-17T15:54:42+00:00"),
                ("time_period.start", "2022-05-17T13:09:58+00:00"),
            ]
        )

        self.assertDictEqual(self.z3d.station_metadata.to_dict(single=True), sm)

    def test_zen_esponse(self):
        self.assertEqual(None, self.z3d.zen_response)
        # self.assertDictEqual(
        #     self.z3d.zen_response.to_dict(single=True),
        #     self.zr.to_dict(single=True),
        # )

    def test_coil_response_filter(self):

        with self.subTest("frequency"):
            self.assertTrue(
                np.isclose(
                    self.z3d.coil_response.frequencies, self.cr.frequencies
                ).all()
            )

        with self.subTest("amplitude"):
            self.assertTrue(
                np.isclose(
                    self.z3d.coil_response.amplitudes, self.cr.amplitudes
                ).all()
            )

        with self.subTest("phase"):
            self.assertTrue(
                np.isclose(self.z3d.coil_response.phases, self.cr.phases).all()
            )

    def test_conversion_filter(self):

        self.assertEqual(
            self.z3d.counts2mv_filter.to_dict(single=True),
            self.cf.to_dict(single=True),
        )

    def channel_response(self):

        cr = ChannelResponseFilter(filters_list=[self.cf, self.cr])

        self.assertListEqual(
            cr.filters_list, self.z3d.channel_response.filters_list
        )


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Z3DCollection
=================

An object to hold Z3D file information to make processing easier.


Created on Sat Apr  4 12:40:40 2020

@author: peacock
"""
# =============================================================================
# Imports
# =============================================================================
import pandas as pd

from mth5.io.collection import Collection
from mth5.io.zen import CoilResponse, Z3D

from mt_metadata.timeseries import Station

# =============================================================================
# Collection of Z3D Files
# =============================================================================


class Z3DCollection(Collection):
    """
    An object to deal with a collection of Z3D files. Metadata and information
    are contained with in Pandas DataFrames for easy searching.

    """

    def __init__(self, file_path=None, **kwargs):

        super().__init__(file_path=file_path, **kwargs)
        self.station_metadata_dict = {}
        self.file_ext = "z3d"

    def get_calibrations(self, antenna_calibration_file):
        """
        Get coil calibrations from the antenna.cal file

        :param antenna_calibration_file: DESCRIPTION
        :type antenna_calibration_file: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """
        return CoilResponse(antenna_calibration_file)

    def _sort_station_metadata(self, station_list):
        """

        :param station_list: DESCRIPTION
        :type station_list: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """
        sdf = pd.DataFrame(station_list)
        info = {}
        for station in sdf.id.unique():
            station_df = sdf[sdf.id == station]
            station_metadata = Station()
            station_metadata.id = station
            station_metadata.location.latitude = station_df[
                "location.latitude"
            ].median()
            station_metadata.location.longitude = station_df[
                "location.longitude"
            ].median()
            station_metadata.location.elevation = station_df[
                "location.elevation"
            ].median()

            info[station] = station_metadata

        return info

    def to_dataframe(
        self, sample_rates=[256, 4096], run_name_zeros=4, calibration_path=None
    ):
        """
        Get general z3d information and put information in a dataframe

        :param z3d_fn_list: List of files Paths to z3d files
        :type z3d_fn_list: list

        :return: Dataframe of z3d information
        :rtype: Pandas.DataFrame

        :Example: ::

            >>> zc_obj = zc.Z3DCollection(r"/home/z3d_files")
            >>> z3d_fn_list = zc.get_z3d_fn_list()
            >>> z3d_df = zc.get_z3d_info(z3d_fn_list)
            >>> # write dataframe to a file to use later
            >>> z3d_df.to_csv(r"/home/z3d_files/z3d_info.csv")

        """

        station_metadata = []
        cal_obj = self.get_calibrations(calibration_path)
        entries = []
        for z3d_fn in self.get_files(
            [self.file_ext, self.file_ext.lower(), self.file_ext.upper()]
        ):
            z3d_obj = Z3D(z3d_fn)
            z3d_obj.read_all_info()
            station_metadata.append(z3d_obj.station_metadata.to_dict(single=True))
            if not int(z3d_obj.sample_rate) in sample_rates:
                self.logger.warning(f"{z3d_obj.sample_rate} not in {sample_rates}")
                return

            entry = {}
            entry["survey"] = z3d_obj.metadata.job_name
            entry["station"] = z3d_obj.station
            entry["run"] = None
            entry["start"] = z3d_obj.start.isoformat()
            entry["end"] = z3d_obj.end.isoformat()
            entry["channel_id"] = z3d_obj.channel_number
            entry["component"] = z3d_obj.component
            entry["fn"] = z3d_fn
            entry["sample_rate"] = z3d_obj.sample_rate
            entry["file_size"] = z3d_obj.file_size
            entry["n_samples"] = z3d_obj.n_samples
            entry["sequence_number"] = 0
            entry["instrument_id"] = f"ZEN_{int(z3d_obj.header.box_number):03}"
            if cal_obj.has_coil_number(z3d_obj.coil_number):
                entry["calibration_fn"] = cal_obj.calibration_file
            else:
                entry["calibration_fn"] = None

            entries.append(entry)
        # make pandas dataframe and set data types
        df = self._sort_df(self._set_df_dtypes(pd.DataFrame(entries)), run_name_zeros)

        self.station_metadata_dict = self._sort_station_metadata(station_metadata)

        return df

    def assign_run_names(self, df, zeros=3):

        # assign run names
        for station in df.station.unique():
            starts = sorted(df[df.station == station].start.unique())
            for block_num, start in enumerate(starts, 1):
                sample_rate = df[
                    (df.station == station) & (df.start == start)
                ].sample_rate.unique()[0]

                df.loc[
                    (df.station == station) & (df.start == start), "run"
                ] = f"sr{sample_rate:.0f}_{block_num:0{zeros}}"
                df.loc[
                    (df.station == station) & (df.start == start),
                    "sequence_number",
                ] = block_num
        return df



# -*- coding: utf-8 -*-
"""
Module to read and parse native Phoenix Geophysics data formats of the MTU-5C
Family.

This module implements Streamed readers for segmented-decimated time series 
formats of the MTU-5C family.

:author: Jorge Torres-Solis

Revised 2022 by J. Peacock 

"""

# =============================================================================
# Imports
# =============================================================================

import numpy as np
from numpy.lib.stride_tricks import as_strided

from struct import unpack_from, unpack
from mth5.io.phoenix.readers import TSReaderBase
from mth5.timeseries import ChannelTS

AD_IN_AD_UNITS = 0
AD_INPUT_VOLTS = 1
INSTRUMENT_INPUT_VOLTS = 2

# =============================================================================


class NativeReader(TSReaderBase):
    """
    Native sampling rate 'Raw' time series reader class, these are the .bin
    files.  They are formatted with a header of 128 bytes then frames of 64.

    Each frame is 20 x 3 byte (24-bit) data point then a 4 byte footer.

    """

    def __init__(
        self,
        path,
        num_files=1,
        scale_to=AD_INPUT_VOLTS,
        header_length=128,
        last_frame=0,
        ad_plus_minus_range=5.0,
        channel_type="E",
        report_hw_sat=False,
        **kwargs,
    ):
        # Init the base class
        super().__init__(
            path, num_files, header_length, report_hw_sat, **kwargs
        )

        self._chunk_size = 4096

        # Track the last frame seen by the streamer, to report missing frames
        self.last_frame = last_frame
        self.header_length = header_length
        self.data_scaling = scale_to
        self.ad_plus_minus_range = ad_plus_minus_range

        if header_length == 128:
            self.unpack_header(self.stream)
        # Now that we have the channel circuit-based gain (either form init
        # or from the header)
        # We can calculate the voltage range at the input of the board.
        self.input_plusminus_range = self._calculate_input_plusminus_range()

        self.scale_factor = self._calculate_data_scaling()

        # optimization variables
        self.footer_idx_samp_mask = int("0x0fffffff", 16)
        self.footer_sat_mask = int("0x70000000", 16)

    def _calculate_input_plusminus_range(self):
        """
        set the correct input plusminus range from metadata
        """
        return self.ad_plus_minus_range / self.total_circuitry_gain

    def _calculate_data_scaling(self):
        """
        Get the correct data scaling for the AD converter
        """
        if self.data_scaling == AD_IN_AD_UNITS:
            return 256
        elif self.data_scaling == AD_INPUT_VOLTS:
            return self.ad_plus_minus_range / (2**31)
        elif self.data_scaling == INSTRUMENT_INPUT_VOLTS:
            return self.input_plusminus_range / (2**31)
        else:
            raise LookupError("Invalid scaling requested")

    def read_frames(self, num_frames):
        """
        Read the given amount of frames from the data.

        .. note:: that seek is not reset so if you iterate this the stream
         reads from the last tell.

        :param num_frames: Number of frames to read
        :type num_frames: integer
        :return: Scaled data from the given number of frames
        :rtype: np.ndarray(dtype=float)

        """

        frames_in_buf = 0
        _idx_buf = 0
        _data_buf = np.empty([num_frames * 20])  # 20 samples packed in a frame

        while frames_in_buf < num_frames:

            dataFrame = self.stream.read(64)
            if not dataFrame:
                if not self.open_next():
                    return np.empty([0])
                dataFrame = self.stream.read(64)
                if not dataFrame:
                    return np.empty([0])
            dataFooter = unpack_from("I", dataFrame, 60)

            # Check that there are no skipped frames
            frameCount = dataFooter[0] & self.footer_idx_samp_mask
            difCount = frameCount - self.last_frame
            if difCount != 1:
                self.logger.warning(
                    "Ch [%s] Missing frames at %d [%d]\n"
                    % (self.channel_id, frameCount, difCount)
                )
            self.last_frame = frameCount

            for ptrSamp in range(0, 60, 3):
                # unpack expects 4 bytes, but the frames are only 3?
                value = unpack(
                    ">i", dataFrame[ptrSamp : ptrSamp + 3] + b"\x00"
                )[0]
                _data_buf[_idx_buf] = value * self.scale_factor
                _idx_buf += 1
            frames_in_buf += 1

            if self.report_hw_sat:
                satCount = (dataFooter[0] & self.footer_sat_mask) >> 24
                if satCount:
                    self.logger.warning(
                        "Ch [%s] Frame %d has %d saturations"
                        % (self.ch_id, frameCount, satCount)
                    )
        return _data_buf

    @property
    def npts_per_frame(self):
        return int((self.frame_size_bytes - 4) / 3)

    def read(self):
        """
        Read the full data file.

        .. note:: This uses :class:`numpy.lib.stride_tricks.as_strided` which
         can be unstable if the bytes are not the correct length.  See notes by
         numpy.

        Got this solution from:
        https://stackoverflow.com/questions/12080279/how-do-i-create-a-numpy-dtype-that-includes-24-bit-integers?msclkid=3398046ecd6511ec9a37394f28c5aaba

        :return: scaled data and footer
        :rtype: tuple (data, footer)

        """

        # should do a memory map otherwise things can go badly with as_strided
        raw_data = np.memmap(self.base_path, ">i1", mode="r")
        raw_data = raw_data[self.header_length :]

        # trim of any extra bytes
        extra_bytes = raw_data.size % 64
        if extra_bytes != 0:
            self.warning(f"found {extra_bytes} extra bits in file.")
        useable_bytes = raw_data.size - extra_bytes
        raw_data = raw_data[:useable_bytes]
        # This should now be the exact number of frames after trimming off
        # extra bytes.
        n_frames = int(raw_data.size / self.frame_size_bytes)

        # reshape to (nframes, 64)
        raw_data = raw_data.reshape((n_frames, self.frame_size_bytes))

        # split the data from the footer
        ts = raw_data[:, 0 : self.npts_per_frame * 3].flatten()
        footer = raw_data[:, self.npts_per_frame * 3 :].flatten()

        # get the number of raw byte frames
        raw_frames = int(ts.size / 12)

        # stride over bytes making new 4 bytes for a 32bit integer and scale
        ts_data = (
            as_strided(
                ts.view(np.int32),
                strides=(12, 3),
                shape=(raw_frames, 4),
            )
            .flatten()
            .byteswap()
            * self.scale_factor
        )
        # somehow the number is off by just a bit ~1E-7 V

        # view the footer as an int32
        footer = footer.view(np.int32)

        return ts_data, footer

    def read_sequence(self, start=0, end=None):
        """
        Read sequence of files into a single array

        :param start: sequence start, defaults to 0
        :type start: integer, optional
        :param end: sequence end, defaults to None
        :type end: integer, optional
        :return: scaled data
        :rtype: np.ndarray(dtype=float32)
        :return: footer
        :rtype: np.ndarray(dtype=int32)

        """

        data = np.array([])
        footer = np.array([])
        for fn in self.sequence_list[slice(start, end)]:
            self._open_file(fn)
            self.unpack_header(self.stream)
            ts, foot = self.read()
            data = np.append(data, ts)
            footer = np.append(footer, foot)
        return data, footer

    def skip_frames(self, num_frames):
        """
        Skip frames of the stream

        :param num_frames: number of frames to skip
        :type num_frames: integer
        :return: end of file
        :rtype: boolean

        """
        bytes_to_skip = int(num_frames * 64)
        # Python is dumb for seek and tell, it cannot tell us if a seek goes
        # past EOF so instead we need to do inefficient reads to skip bytes
        while bytes_to_skip > 0:
            foo = self.stream.read(bytes_to_skip)
            local_read_size = len(foo)

            # If we ran out of data in this file before finishing the skip,
            # open the next file and return false if there is no next file
            # to indicate that the skip ran out of
            # data before completion
            if local_read_size == 0:
                more_data = self.open_next()
                if not more_data:
                    return False
            else:
                bytes_to_skip -= local_read_size
        # If we reached here we managed to skip all the data requested
        # return true
        self.last_frame += num_frames
        return True

    def to_channel_ts(self):
        """
        convert to a ChannelTS object

        :return: DESCRIPTION
        :rtype: TYPE

        """
        data, footer = self.read()
        ch_metadata = self.channel_metadata()
        return ChannelTS(
            channel_type=ch_metadata.type,
            data=data,
            channel_metadata=ch_metadata,
            run_metadata=self.run_metadata(),
            station_metadata=self.station_metadata(),
        )



# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 10:56:45 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest

from mth5.io.nims import GPS, GPSError

# =============================================================================

gprmc_string = (
    "GPRMC,183511,A,3443.6098,N,11544.1007,W,000.0,000.0,260919,013.1,E"
)
gpgga_string = (
    "GPGGA,183511,3443.6098,N,11544.1007,W,1,04,2.6,937.2,M,-28.1,M,,"
)


class TestGPS(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.gprmc_string = "GPRMC,183511,A,3443.6098,N,11544.1007,W,000.0,000.0,260919,013.1,E*"
        self.gpgga_string = (
            "GPGGA,183511,3443.6098,N,11544.1007,W,1,04,2.6,937.2,M,-28.1,M,*"
        )
        self.gprmc_bytes = self.gprmc_string.encode()
        self.gpgga_bytes = self.gpgga_string.encode()

        self.gps_obj = GPS("")

        self.a_list = [
            self.gprmc_string,
            self.gpgga_string,
            self.gprmc_string,
            self.gpgga_string,
        ]

        self.b_list = [
            self.gprmc_string,
            self.gpgga_string,
            self.gprmc_bytes,
            self.gpgga_bytes,
        ]

    def test_validate_string(self):

        for aa, bb in zip(self.a_list, self.b_list):

            with self.subTest(aa[0:5]):
                self.assertEqual(
                    aa[:-1],
                    self.gps_obj.validate_gps_string(bb),
                )

    def test_split_list(self):
        for aa, bb in zip(self.a_list, self.b_list):

            with self.subTest(aa[0:5]):
                self.assertListEqual(
                    aa[:-1].split(","),
                    self.gps_obj._split_gps_string(bb),
                )

    def test_validate_list(self):
        for aa, bb in zip(self.a_list, self.b_list):
            input_list = self.gps_obj._split_gps_string(bb)
            gps_list, error_list = self.gps_obj.validate_gps_list(input_list)
            with self.subTest(aa[0:5]):

                self.assertListEqual(aa[:-1].split(","), gps_list)

            with self.subTest("error list"):
                self.assertListEqual([], error_list)

    def test_validate_latitude_fail(self):
        with self.subTest("length"):
            self.assertRaises(
                GPSError, self.gps_obj._validate_latitude, "3443.60", "N"
            )

        with self.subTest("hemisphere length"):
            self.assertRaises(
                GPSError, self.gps_obj._validate_latitude, "3443.6098", "NS"
            )

        with self.subTest("hemisphere"):
            self.assertRaises(
                GPSError, self.gps_obj._validate_latitude, "3443.6098", "Z"
            )

    def test_validate_longitude_fail(self):
        with self.subTest("length"):
            self.assertRaises(
                GPSError, self.gps_obj._validate_longitude, "115.17", "W"
            )

        with self.subTest("hemisphere length"):
            self.assertRaises(
                GPSError, self.gps_obj._validate_longitude, "11544.1007", "WE"
            )

        with self.subTest("hemisphere"):
            self.assertRaises(
                GPSError, self.gps_obj._validate_longitude, "11544.1007", "T"
            )

    def test_validate_elevation_fail(self):
        self.assertRaises(GPSError, self.gps_obj._validate_elevation, "kz")

    def test_validate_time_fail(self):
        with self.subTest("legth"):
            self.assertRaises(GPSError, self.gps_obj._validate_time, "0115")

        with self.subTest("type"):
            self.assertRaises(GPSError, self.gps_obj._validate_time, "0115SA")

    def test_validate_date_fail(self):
        with self.subTest("legth"):
            self.assertRaises(GPSError, self.gps_obj._validate_date, "0115")

        with self.subTest("type"):
            self.assertRaises(GPSError, self.gps_obj._validate_date, "0115SA")


class TestGPRMC(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.gprmc_string = "GPRMC,183511,A,3443.6098,N,11544.1007,W,000.0,000.0,260919,013.1,E*"
        self.gps_obj = GPS(self.gprmc_string)

    def test_type(self):
        self.assertEqual(self.gps_obj.gps_type, "GPRMC")

    def test_latitude(self):
        self.assertAlmostEqual(self.gps_obj.latitude, 34.72683)

    def test_longitude(self):
        self.assertAlmostEqual(self.gps_obj.longitude, -115.73501166666667)

    def test_elevation(self):
        self.assertAlmostEqual(self.gps_obj.elevation, 0)

    def test_declination(self):
        self.assertAlmostEqual(self.gps_obj.declination, 13.1)

    def test_time_stamp(self):
        self.assertEqual(
            self.gps_obj.time_stamp.isoformat(), "2019-09-26T18:35:11"
        )


class TestGPGGA(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.gpgga_string = (
            "GPGGA,183511,3443.6098,N,11544.1007,W,1,04,2.6,937.2,M,-28.1,M,*"
        )
        self.gps_obj = GPS(self.gpgga_string)

    def test_type(self):
        self.assertEqual(self.gps_obj.gps_type, "GPGGA")

    def test_latitude(self):
        self.assertAlmostEqual(self.gps_obj.latitude, 34.72683)

    def test_longitude(self):
        self.assertAlmostEqual(self.gps_obj.longitude, -115.73501166666667)

    def test_elevation(self):
        self.assertAlmostEqual(self.gps_obj.elevation, 937.2)

    def test_declination(self):
        self.assertEqual(self.gps_obj.declination, None)

    def test_time_stamp(self):
        self.assertEqual(
            self.gps_obj.time_stamp.isoformat(), "1980-01-01T18:35:11"
        )


# =============================================================================
# Run
# =============================================================================
if __name__ in "__main__":
    unittest.main()



"""
Module to read and parse native Phoenix Geophysics data formats of the 
MTU-5C Family.

This module implements Streamed readers for segmented-decimated continuus-decimated
and native sampling rate time series formats of the MTU-5C family.

:author: Jorge Torres-Solis

Revised 2022 by J. Peacock 
"""

# =============================================================================
# Imports
# =============================================================================
from pathlib import Path
from .header import Header

from mth5.utils.mth5_logger import setup_logger

# =============================================================================


class TSReaderBase(Header):
    """

    Generic reader that all other readers will inherit

    """

    def __init__(
        self,
        path,
        num_files=1,
        header_length=128,
        report_hw_sat=False,
        **kwargs,
    ):
        self._seq = None
        super().__init__(
            header_length=header_length, report_hw_sat=report_hw_sat, **kwargs
        )

        self.logger = setup_logger(
            f"{self.__class__}.{self.__class__.__name__}"
        )
        self.base_path = path
        self.last_seq = self.seq + num_files
        self.stream = None
        # Open the file passed as the first file in the sequence to stream
        self._open_file(self.base_path)
        if self._recording_id is None:
            self.recording_id = self.base_path.stem.split("_")[1]
        if self._channel_id is None:
            self.channel_id = self.base_path.stem.split("_")[2]

    @property
    def base_path(self):
        """

        :return: full path of file
        :rtype: :class:`pathlib.Path`

        """
        return self._base_path

    @base_path.setter
    def base_path(self, value):
        """

        :param value: full path to file
        :type value: string or :class:`pathlib.Path`

        """

        self._base_path = Path(value)

    @property
    def base_dir(self):
        """

        :return: parent directory of file
        :rtype: :class:`pathlib.Path`

        """
        return self.base_path.parent

    @property
    def file_name(self):
        """

        :return: name of the file
        :rtype: string

        """
        return self.base_path.name

    @property
    def file_extension(self):
        """

        :return: file extension
        :rtype: string

        """
        return self.base_path.suffix

    @property
    def instrument_id(self):
        """

        :return: instrument ID
        :rtype: string

        """
        return self.base_path.stem.split("_")[0]

    @property
    def seq(self):
        """

        :return: sequence number of the file
        :rtype: int

        """
        if self._seq is None:
            return int(self.base_path.stem.split("_")[3], 16)
        return self._seq

    @seq.setter
    def seq(self, value):
        """

        :param value: sequence number
        :type value: integer


        """
        self._seq = int(value)

    @property
    def file_size(self):
        """

        :return: file size in bytes
        :rtype: integer

        """
        return self.base_path.stat().st_size

    @property
    def max_samples(self):
        """
        Max number of samples in a file which is:

        (total number of bytes - header length) / frame size * n samples per frame

        :return: max number of samples in a file
        :rtype: int

        """
        return int((self.file_size - self.header_length) / 4)

    @property
    def sequence_list(self):
        """
        get all the files in the sequence sorted by sequence number
        """
        return sorted(list(self.base_dir.glob(f"*{self.file_extension}")))

    def _open_file(self, filename):
        """
        open a given file in 'rb' mode

        :param filename: full path to file
        :type filename: :class:`pathlib.Path`
        :return: boolean if the file is now open [True] or not [False]
        :rtype: boolean

        """
        filename = Path(filename)

        if filename.exists():
            self.logger.debug(f"Opening {filename}")
            self.stream = open(filename, "rb")
            return True
        return False

    def open_next(self):
        """
        Open the next file in the sequence
        :return: [True] if next file is now open, [False] if it is not
        :rtype: boolean

        """
        if self.stream is not None:
            self.stream.close()
        self.seq += 1
        self.open_file_seq(self.seq)
        if self.seq < self.last_seq:
            new_path = self.sequence_list[self.seq - 1]
            return self._open_file(new_path)
        return False

    def open_file_seq(self, file_seq_num=None):
        """
        Open a file in the sequence given the sequence number
        :param file_seq_num: sequence number to open, defaults to None
        :type file_seq_num: integer, optional
        :return: [True] if next file is now open, [False] if it is not
        :rtype: boolean

        """
        if self.stream is not None:
            self.stream.close()
        if file_seq_num is not None:
            self.seq = file_seq_num
        new_path = self.sequence_list[self.seq - 1]
        return self._open_file(new_path)

    def close(self):
        """
        Close the file

        """
        if self.stream is not None:
            self.stream.close()



"""
Logger
===========

Setup a logger with two handlers to remove redundancy between logs entries
One is a stream handler for any messages to the console.  The other is
either a file handler or a null handler.

"""

from pathlib import Path
import yaml
import logging
import logging.config
from concurrent_log_handler import ConcurrentRotatingFileHandler

# =============================================================================
# Global Variables
# =============================================================================
LEVEL_DICT = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

LOG_FORMAT = logging.Formatter(
    "%(asctime)s [line %(lineno)d] %(name)s.%(funcName)s - %(levelname)s: %(message)s"
)
# Get the configuration file path, should be in same directory as this file
CONF_PATH = Path(__file__).parent
CONF_FILE = Path.joinpath(CONF_PATH, "logging_config.yaml")

# make a folder for the logs to go into.
LOG_PATH = CONF_PATH.parent.parent.joinpath("logs")

if not LOG_PATH.exists():
    LOG_PATH.mkdir()

if not CONF_FILE.exists():
    CONF_FILE = None


def load_logging_config(config_fn=CONF_FILE):
    """
    configure/setup the logging according to the input configfile

    :param configfile: .yml, .ini, .conf, .json, .yaml.
    Its default is the logging.yml located in the same dir as this module.
    It can be modofied to use env variables to search for a log config file.
    """
    if config_fn:
        config_file = Path(config_fn)
        with open(config_file, "r") as fid:
            config_dict = yaml.safe_load(fid)

    else:
        config_dict = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "standard": {
                    "format": "%(asctime)s [line %(lineno)d] %(name)s.%(funcName)s - %(levelname)s: %(message)s",
                    "datefmt": "%Y-%m-%dT%H:%M:%S",
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "INFO",
                    "formatter": "standard",
                    "stream": "ext://sys.stdout",
                }
            },
            "loggers": {
                "__main__": {
                    "level": "DEBUG",
                    "handlers": ["console"],
                    "propagate": False,
                }
            },
            "root": {"level": "DEBUG", "handlers": ["console"], "propogate": False},
        }

    logging.config.dictConfig(config_dict)


def setup_logger(logger_name, fn=None, level="debug"):
    """
    Create a logger, can write to a separate file.  This will write to
    the logs folder in the mt_metadata directory.

    :param logger_name: name of the logger, typically __name__
    :type logger_name: string
    :param fn: file name to write to, defaults to None
    :type fn: TYPE, optional
    :param level: DESCRIPTION, defaults to "debug"
    :type level: TYPE, optional
    :return: DESCRIPTION
    :rtype: TYPE

    """

    logger = logging.getLogger(logger_name)

    # if there is a file name create file in logs directory
    if fn is not None:
        # need to clear the handlers to make sure there is only
        # one call per logger plus stdout
        if logger.hasHandlers():
            logger.handlers.clear()

        logger.propagate = False
        # want to add a stream handler for any Info print statements as stdOut
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(LOG_FORMAT)
        stream_handler.setLevel(LEVEL_DICT["info"])
        logger.addHandler(stream_handler)

        fn = LOG_PATH.joinpath(fn)
        exists = False
        if fn.exists():
            exists = True

        if fn.suffix not in [".log"]:
            fn = Path(fn.parent, f"{fn.stem}.log")

        # fn_handler = logging.FileHandler(fn)
        fn_handler = ConcurrentRotatingFileHandler(fn, maxBytes=2 ** 21, backupCount=2)
        fn_handler.setFormatter(LOG_FORMAT)
        fn_handler.setLevel(LEVEL_DICT[level.lower()])
        logger.addHandler(fn_handler)
        if not exists:
            logger.info(f"Logging file can be found {logger.handlers[-1].baseFilename}")

    return logger



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  1 12:57:32 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
from pathlib import Path
import dateutil
import logging

from mt_metadata.utils.mttime import MTime

# =============================================================================
class NIMSError(Exception):
    pass


class NIMSHeader(object):
    """
    class to hold the NIMS header information.

    A typical header looks like

    .. code-block::

        '''
        >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        >>>user field>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        SITE NAME: Budwieser Spring
        STATE/PROVINCE: CA
        COUNTRY: USA
        >>> The following code in double quotes is REQUIRED to start the NIMS <<
        >>> The next 3 lines contain values required for processing <<<<<<<<<<<<
        >>> The lines after that are optional <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        "300b"  <-- 2CHAR EXPERIMENT CODE + 3 CHAR SITE CODE + RUN LETTER
        1105-3; 1305-3  <-- SYSTEM BOX I.D.; MAG HEAD ID (if different)
        106  0 <-- N-S Ex WIRE LENGTH (m); HEADING (deg E mag N)
        109  90 <-- E-W Ey WIRE LENGTH (m); HEADING (deg E mag N)
        1         <-- N ELECTRODE ID
        3         <-- E ELECTRODE ID
        2         <-- S ELECTRODE ID
        4         <-- W ELECTRODE ID
        Cu        <-- GROUND ELECTRODE INFO
        GPS INFO: 01/10/19 16:16:42 1616.7000 3443.6088 115.7350 W 946.6
        OPERATOR: KP
        COMMENT: N/S CRS: .95/.96 DCV: 3.5 ACV:1
        E/W CRS: .85/.86 DCV: 1.5 ACV: 1
        Redeployed site for run b b/c possible animal disturbance
        '''

    """

    def __init__(self, fn=None):
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}"
        )
        self.fn = fn
        self._max_header_length = 1000
        self.header_dict = None
        self.site_name = None
        self.state_province = None
        self.country = None
        self.box_id = None
        self.mag_id = None
        self.ex_length = None
        self.ex_azimuth = None
        self.ey_length = None
        self.ey_azimuth = None
        self.n_electrode_id = None
        self.s_electrode_id = None
        self.e_electrode_id = None
        self.w_electrode_id = None
        self.ground_electrode_info = None
        self.header_gps_stamp = None
        self.header_gps_longitude = None
        self.header_gps_longitude = None
        self.header_gps_elevation = None
        self.operator = None
        self.comments = None
        self.run_id = None
        self.data_start_seek = 0

    @property
    def fn(self):
        """Full path to NIMS file"""
        return self._fn

    @fn.setter
    def fn(self, value):
        if value is not None:
            self._fn = Path(value)
        else:
            self._fn = None

    @property
    def station(self):
        """Station ID"""
        if self.run_id is not None:
            return self.run_id[0:-1]

    @property
    def file_size(self):
        """Size of the file"""
        if self.fn is not None:
            return self.fn.stat().st_size

    def read_header(self, fn=None):
        """
        read header information

        :param fn: full path to file to read
        :type fn: string or :class:`pathlib.Path`
        :raises: :class:`mth5.io.nims.NIMSError` if something is not right.

        """
        if fn is not None:
            self.fn = fn

        if not self.fn.exists():
            msg = f"Could not find nims file {self.fn}"
            self.logger.error(msg)
            raise NIMSError(msg)

        self.logger.debug(f"Reading NIMS file {self.fn}")

        ### load in the entire file, its not too big
        with open(self.fn, "rb") as fid:
            header_str = fid.read(self._max_header_length)
            header_list = header_str.split(b"\r")

        self.header_dict = {}
        last_index = len(header_list)
        last_line = header_list[-1]
        for ii, line in enumerate(header_list[0:-1]):
            if ii == last_index:
                break
            if b"comments" in line.lower():
                last_line = header_list[ii + 1]
                last_index = ii + 1

            line = line.decode()
            if line.find(">") == 0:
                continue
            elif line.find(":") > 0:
                key, value = line.split(":", 1)
                self.header_dict[key.strip().lower()] = value.strip()
            elif line.find("<--") > 0:
                value, key = line.split("<--")
                self.header_dict[key.strip().lower()] = value.strip()
        ### sometimes there are some spaces before the data starts
        if last_line.count(b" ") > 0:
            if last_line[0:1] == b" ":
                last_line = last_line.strip()
            else:
                last_line = last_line.split()[1].strip()
        data_start_byte = last_line[0:1]
        ### sometimes there are rogue $ around
        if data_start_byte in [b"$", b"g"]:
            data_start_byte = last_line[1:2]
        self.data_start_seek = header_str.find(data_start_byte)

        self.parse_header_dict()

    def parse_header_dict(self, header_dict=None):
        """
        parse the header dictionary into something useful
        """
        if header_dict is not None:
            self.header_dict = header_dict

        assert isinstance(self.header_dict, dict)

        for key, value in self.header_dict.items():
            if "wire" in key:
                if key.find("n") == 0:
                    self.ex_length = float(value.split()[0])
                    self.ex_azimuth = float(value.split()[1])
                elif key.find("e") == 0:
                    self.ey_length = float(value.split()[0])
                    self.ey_azimuth = float(value.split()[1])
            elif "system" in key:
                self.box_id = value.split(";")[0].strip()
                self.mag_id = value.split(";")[1].strip()
            elif "gps" in key:
                gps_list = value.split()
                self.header_gps_stamp = MTime(
                    dateutil.parser.parse(
                        " ".join(gps_list[0:2]), dayfirst=True
                    )
                )
                self.header_gps_latitude = self._get_latitude(
                    gps_list[2], gps_list[3]
                )
                self.header_gps_longitude = self._get_longitude(
                    gps_list[4], gps_list[5]
                )
                self.header_gps_elevation = float(gps_list[6])
            elif "run" in key:
                self.run_id = value.replace('"', "")
            else:
                setattr(self, key.replace(" ", "_").replace("/", "_"), value)

    def _get_latitude(self, latitude, hemisphere):
        """
        Get latitude as decimal degrees

        :param latitude: latitude
        :type latitude: float
        :param hemisphere: hemisphere id [ 'N' | 'S' ]
        :type hemisphere: string
        :return: latitude in decimal degrees with proper sign
        :rtype: float

        """
        if not isinstance(latitude, float):
            latitude = float(latitude)
        if hemisphere.lower() == "n":
            return latitude
        if hemisphere.lower() == "s":
            return -1 * latitude

    def _get_longitude(self, longitude, hemisphere):
        """
        Get longitude as decimal degrees

        :param longitude: longitude
        :type longitude: float
        :param hemisphere: hemisphere id [ 'N' | 'S' ]
        :type hemisphere: string
        :return: latitude in decimal degrees with proper sign
        :rtype: float
        """
        if not isinstance(longitude, float):
            longitude = float(longitude)
        if hemisphere.lower() == "e":
            return longitude
        if hemisphere.lower() == "w":
            return -1 * longitude



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  1 11:43:56 2022

@author: jpeacock
"""
# =============================================================================
# Imports
# =============================================================================
import dateutil
import logging


# =============================================================================
class GPSError(Exception):
    pass


class GPS(object):
    """
    class to parse GPS stamp from the NIMS

    Depending on the type of Stamp different attributes will be filled.

    GPRMC has full date and time information and declination
    GPGGA has elevation data

    .. note:: GPGGA date is set to 1980-01-01 so that the time can be estimated.
              Should use GPRMC for accurate date/time information.
    """

    def __init__(self, gps_string, index=0):

        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}"
        )

        self.gps_string = gps_string
        self.index = index
        self._type = None
        self._time = None
        self._date = "010180"
        self._latitude = None
        self._latitude_hemisphere = None
        self._longitude = None
        self._longitude_hemisphere = None
        self._declination = None
        self._declination_hemisphere = None
        self._elevation = None
        self.valid = False
        self.elevation_units = "meters"

        self.type_dict = {
            "gprmc": {
                0: "type",
                1: "time",
                2: "fix",
                3: "latitude",
                4: "latitude_hemisphere",
                5: "longitude",
                6: "longitude_hemisphere",
                7: "skip",
                8: "skip",
                9: "date",
                10: "declination",
                11: "declination_hemisphere",
                "length": [12],
                "type": 0,
                "time": 1,
                "fix": 2,
                "latitude": 3,
                "latitude_hemisphere": 4,
                "longitude": 5,
                "longitude_hemisphere": 6,
                "date": 9,
                "declination": 10,
            },
            "gpgga": {
                0: "type",
                1: "time",
                2: "latitude",
                3: "latitude_hemisphere",
                4: "longitude",
                5: "longitude_hemisphere",
                6: "var_01",
                7: "var_02",
                8: "var_03",
                9: "elevation",
                10: "elevation_units",
                11: "elevation_error",
                12: "elevation_error_units",
                13: "null_01",
                14: "null_02",
                "length": [14, 15],
                "type": 0,
                "time": 1,
                "latitude": 2,
                "latitude_hemisphere": 3,
                "longitude": 4,
                "longitude_hemisphere": 5,
                "elevation": 9,
                "elevation_units": 10,
                "elevation_error": 11,
                "elevation_error_units": 12,
            },
        }
        self.parse_gps_string(self.gps_string)

    def __str__(self):
        """string representation"""
        msg = [
            f"type = {self.gps_type}",
            f"index = {self.index}",
            f"fix = {self.fix}",
            f"time_stamp =  {self.time_stamp}",
            f"latitude = {self.latitude}",
            f"longitude = {self.longitude}",
            f"elevation = {self.elevation}",
            f"declination = {self.declination}",
        ]

        return "\n".join(msg)

    def __repr__(self):
        return self.__str__()

    def validate_gps_string(self, gps_string):
        """
        make sure the string is valid, remove any binary numbers and find
        the end of the string as '*'

        :param string gps_string: raw GPS string to be validated
        :returns: validated string or None if there is something wrong

        """

        if isinstance(gps_string, bytes):
            for replace_str in [b"\xd9", b"\xc7", b"\xcc"]:
                gps_string = gps_string.replace(replace_str, b"")

            ### sometimes the end is set with a zero for some reason
            gps_string = gps_string.replace(b"\x00", b"*")

            if gps_string.find(b"*") < 0:
                logging.debug(f"GPSError: No end to stamp {gps_string}")
                return None
            else:
                try:
                    gps_string = gps_string[0 : gps_string.find(b"*")].decode()
                    return gps_string
                except UnicodeDecodeError:
                    logging.debug(
                        f"GPSError: stamp not correct format, {gps_string}"
                    )
                    return None
        elif isinstance(gps_string, str):
            if "*" not in gps_string:
                logging.debug(f"GPSError: No end to stamp {gps_string}")
                return None

            return gps_string[0 : gps_string.find("*")]

        else:
            raise TypeError(
                f"input must be a string or bytes object, not {type(gps_string)}"
            )

    def _split_gps_string(self, gps_string, delimiter=","):
        """Split a GPS string by ',' and validate it"""

        gps_string = self.validate_gps_string(gps_string)
        if gps_string is None:
            self.valid = False
            return []
        return gps_string.strip().split(",")

    def parse_gps_string(self, gps_string):
        """
        Parse a raw gps string from the NIMS and set appropriate attributes.
        GPS string will first be validated, then parsed.

        :param string gps_string: raw GPS string to be parsed
        """

        gps_list = self._split_gps_string(gps_string)
        if gps_list == []:
            self.logger.debug(f"GPS string is invalid, {gps_string}")
            return

        if len(gps_list) > 1:
            if len(gps_list[1]) > 6:
                self.logger.debug(
                    "GPS time and lat missing a comma adding one, check time"
                )
                gps_list = (
                    gps_list[0:1]
                    + [gps_list[1][0:6], gps_list[1][6:]]
                    + gps_list[2:]
                )

        ### validate the gps list to make sure it is usable
        gps_list, error_list = self.validate_gps_list(gps_list)
        if len(error_list) > 0:
            for error in error_list:
                logging.debug("GPSError: " + error)
        if gps_list is None:
            return

        attr_dict = self.type_dict[gps_list[0].lower()]

        for index, value in enumerate(gps_list):
            setattr(self, "_" + attr_dict[index], value)

        if None not in gps_list:
            self.valid = True
            self.gps_string = gps_string

    def validate_gps_list(self, gps_list):
        """
        check to make sure the gps stamp is the correct format, checks each element
        for the proper format

        :param gps_list: a parsed gps string from a NIMS
        :type gps_list: list
        :raises: :class:`mth5.io.nims.GPSError` if anything is wrong.
        """
        error_list = []
        try:
            gps_list = self._validate_gps_type(gps_list)
        except GPSError as error:
            error_list.append(error.args[0])
            return None, error_list

        ### get the string type
        g_type = gps_list[0].lower()

        ### first check the length, if it is not the proper length then
        ### return, cause you never know if everything else is correct
        try:
            self._validate_list_length(gps_list)
        except GPSError as error:
            error_list.append(error.args[0])
            return None, error_list

        try:
            gps_list[self.type_dict[g_type]["time"]] = self._validate_time(
                gps_list[self.type_dict[g_type]["time"]]
            )
        except GPSError as error:
            error_list.append(error.args[0])
            gps_list[self.type_dict[g_type]["time"]] = None

        try:
            gps_list[
                self.type_dict[g_type]["latitude"]
            ] = self._validate_latitude(
                gps_list[self.type_dict[g_type]["latitude"]],
                gps_list[self.type_dict[g_type]["latitude_hemisphere"]],
            )
        except GPSError as error:
            error_list.append(error.args[0])
            gps_list[self.type_dict[g_type]["latitude"]] = None

        try:
            gps_list[
                self.type_dict[g_type]["longitude"]
            ] = self._validate_longitude(
                gps_list[self.type_dict[g_type]["longitude"]],
                gps_list[self.type_dict[g_type]["longitude_hemisphere"]],
            )
        except GPSError as error:
            error_list.append(error.args[0])
            gps_list[self.type_dict[g_type]["longitude"]] = None

        if g_type == "gprmc":
            try:
                gps_list[
                    self.type_dict["gprmc"]["date"]
                ] = self._validate_date(
                    gps_list[self.type_dict["gprmc"]["date"]]
                )
            except GPSError as error:
                error_list.append(error.args[0])
                gps_list[self.type_dict[g_type]["date"]] = None

        elif g_type == "gpgga":
            try:
                gps_list[
                    self.type_dict["gpgga"]["elevation"]
                ] = self._validate_elevation(
                    gps_list[self.type_dict["gpgga"]["elevation"]]
                )
            except GPSError as error:
                error_list.append(error.args[0])
                gps_list[self.type_dict["gpgga"]["elevation"]] = None

        return gps_list, error_list

    def _validate_gps_type(self, gps_list):
        """Validate gps type should be gpgga or gprmc"""
        gps_type = gps_list[0].lower()
        if "gpg" in gps_type:
            if len(gps_type) > 5:
                gps_list = ["GPGGA", gps_type[-6:]] + gps_list[1:]
            elif len(gps_type) < 5:
                gps_list[0] = "GPGGA"
        elif "gpr" in gps_type:
            if len(gps_type) > 5:
                gps_list = ["GPRMC", gps_type[-6:]] + gps_list[1:]
            elif len(gps_type) < 5:
                gps_list[0] = "GPRMC"

        gps_type = gps_list[0].lower()
        if gps_type not in ["gpgga", "gprmc"]:
            raise GPSError(
                "GPS String type not correct.  "
                f"Expect GPGGA or GPRMC, got {gps_type.upper()}"
            )

        return gps_list

    def _validate_list_length(self, gps_list):
        """validate gps list length based on type of string"""

        gps_list_type = gps_list[0].lower()
        expected_len = self.type_dict[gps_list_type]["length"]
        if len(gps_list) not in expected_len:
            raise GPSError(
                f"GPS string not correct length for {gps_list_type.upper()}.  "
                f"Expected {expected_len}, got {len(gps_list)} "
                f"{','.join(gps_list)}"
            )

    def _validate_time(self, time_str):
        """validate time string, should be 6 characters long and an int"""
        if len(time_str) != 6:
            raise GPSError(
                f"Length of time string {time_str} not correct.  "
                f"Expected 6 got {len(time_str)}. string = {time_str}"
            )
        try:
            int(time_str)
        except ValueError:
            raise GPSError(f"Could not convert time string {time_str}")

        return time_str

    def _validate_date(self, date_str):
        """validate date string, should be 6 characters long and an int"""
        if len(date_str) != 6:
            raise GPSError(
                f"Length of date string not correct {date_str}.  "
                f"Expected 6 got {len(date_str)}. string = {date_str}"
            )
        try:
            int(date_str)
        except ValueError:
            raise GPSError(f"Could not convert date string {date_str}")

        return date_str

    def _validate_latitude(self, latitude_str, hemisphere_str):
        """validate latitude, should have hemisphere string with it"""

        if len(latitude_str) < 8:
            raise GPSError(
                f"Latitude string should be larger than 7 characters.  "
                f"Got {len(latitude_str)}. string = {latitude_str}"
            )
        if len(hemisphere_str) != 1:
            raise GPSError(
                "Latitude hemisphere should be 1 character.  "
                f"Got {len(hemisphere_str)}. string = {hemisphere_str}"
            )
        if hemisphere_str.lower() not in ["n", "s"]:
            raise GPSError(
                f"Latitude hemisphere {hemisphere_str.upper()} not understood"
            )
        try:
            float(latitude_str)
        except ValueError:
            raise GPSError(f"Could not convert latitude string {latitude_str}")

        return latitude_str

    def _validate_longitude(self, longitude_str, hemisphere_str):
        """validate longitude, should have hemisphere string with it"""

        if len(longitude_str) < 8:
            raise GPSError(
                "Longitude string should be larger than 7 characters.  "
                f"Got {len(longitude_str)}. string = {longitude_str}"
            )
        if len(hemisphere_str) != 1:
            raise GPSError(
                "Longitude hemisphere should be 1 character.  "
                f"Got {len(hemisphere_str)}. string = {hemisphere_str}"
            )
        if hemisphere_str.lower() not in ["e", "w"]:
            raise GPSError(
                f"Longitude hemisphere {hemisphere_str.upper()} not understood"
            )
        try:
            float(longitude_str)
        except ValueError:
            raise GPSError(
                f"Could not convert longitude string {longitude_str}"
            )

        return longitude_str

    def _validate_elevation(self, elevation_str):
        """validate elevation, check for converstion to float"""
        elevation_str = elevation_str.lower().replace("m", "")
        if elevation_str == "":
            elevation_str = "0"
        try:
            elevation_str = f"{float(elevation_str)}"
        except ValueError:
            raise GPSError(f"Elevation could not be converted {elevation_str}")

        return elevation_str

    @property
    def latitude(self):
        """
        Latitude in decimal degrees, WGS84
        """
        if (
            self._latitude is not None
            and self._latitude_hemisphere is not None
        ):
            index = len(self._latitude) - 7
            lat = (
                float(self._latitude[0:index])
                + float(self._latitude[index:]) / 60
            )
            if "s" in self._latitude_hemisphere.lower():
                lat *= -1
            return lat
        return 0.0

    @property
    def longitude(self):
        """
        Latitude in decimal degrees, WGS84
        """
        if (
            self._longitude is not None
            and self._longitude_hemisphere is not None
        ):
            index = len(self._longitude) - 7
            lon = (
                float(self._longitude[0:index])
                + float(self._longitude[index:]) / 60
            )
            if "w" in self._longitude_hemisphere.lower():
                lon *= -1
            return lon
        return 0.0

    @property
    def elevation(self):
        """
        elevation in meters
        """
        if self._elevation is not None:
            try:
                return float(self._elevation)
            except ValueError:
                self.logger.error(
                    "GPSError: Could not get elevation GPS string"
                    + f"not complete {self.gps_string}"
                )
        return 0.0

    @property
    def time_stamp(self):
        """
        return a datetime object of the time stamp
        """
        if self._time is None:
            return None
        if self._date is None:
            self._date = "010180"
        try:
            return dateutil.parser.parse(
                "{0} {1}".format(self._date, self._time), dayfirst=True
            )
        except ValueError:
            self.logger.error(f"GPSError: bad date string {self.gps_string}")
            return None

    @property
    def declination(self):
        """
        geomagnetic declination in degrees from north
        """
        if self._declination is None or self._declination_hemisphere is None:
            return None

        dec = float(self._declination)
        if "w" in self._declination_hemisphere.lower():
            dec *= -1
        return dec

    @property
    def gps_type(self):
        """GPRMC or GPGGA"""
        return self._type

    @property
    def fix(self):
        """
        GPS fixed
        """
        if hasattr(self, "_fix"):
            return self._fix
        return None



# -*- coding: utf-8 -*-
"""
====================
Zen
====================

    * Tools for reading and writing files for Zen and processing software
    * Tools for copying data from SD cards
    * Tools for copying schedules to SD cards
    
Created on Tue Jun 11 10:53:23 2013
Updated August 2020 (JP)

:copyright:
    Jared Peacock (jpeacock@usgs.gov)

:license:
    MIT

"""

# ==============================================================================

import datetime
import struct
from pathlib import Path
import logging

import numpy as np

from mt_metadata.utils.mttime import MTime
from mt_metadata.timeseries.filters import (
    ChannelResponseFilter,
    FrequencyResponseTableFilter,
    CoefficientFilter,
)
from mt_metadata.timeseries import Station, Run, Electric, Magnetic

from mth5.io.zen import Z3DHeader, Z3DSchedule, Z3DMetadata
from mth5.timeseries import ChannelTS

# ==============================================================================
#
# ==============================================================================
class Z3D:
    """
    Deals with the raw Z3D files output by zen.
    Arguments
    -----------
        **fn** : string
                 full path to .Z3D file to be read in
    ======================== ================================ =================
    Attributes               Description                      Default Value
    ======================== ================================ =================
    _block_len               length of data block to read in  65536
                             as chunks faster reading
    _counts_to_mv_conversion conversion factor to convert     9.53674316406e-10
                             counts to mv
    _gps_bytes               number of bytes for a gps stamp  16
    _gps_dtype               data type for a gps stamp        see below
    _gps_epoch               starting date of GPS time
                             format is a tuple                (1980, 1, 6, 0,
                                                               0, 0, -1, -1, 0)
    _gps_f0                  first gps flag in raw binary
    _gps_f1                  second gps flag in raw binary
    _gps_flag_0              first gps flag as an int32       2147483647
    _gps_flag_1              second gps flag as an int32      -2147483648
    _gps_stamp_length        bit length of gps stamp          64
    _leap_seconds            leap seconds, difference         16
                             between UTC time and GPS
                             time.  GPS time is ahead
                             by this much
    _week_len                week length in seconds           604800
    df                       sampling rate of the data        256
    fn                       Z3D file name                    None
    gps_flag                 full gps flag                    _gps_f0+_gps_f1
    gps_stamps               np.ndarray of gps stamps         None
    header                   Z3DHeader object                 Z3DHeader
    metadata                 Z3DMetadata                      Z3DMetadata
    schedule                 Z3DSchedule                      Z3DSchedule
    time_series              np.ndarra(len_data)              None
    units                    units in which the data is in    counts
    zen_schedule             time when zen was set to         None
                             run
    ======================== ================================ =================

    * gps_dtype is formated as np.dtype([('flag0', np.int32),
                                        ('flag1', np.int32),
                                        ('time', np.int32),
                                        ('lat', np.float64),
                                        ('lon', np.float64),
                                        ('num_sat', np.int32),
                                        ('gps_sens', np.int32),
                                        ('temperature', np.float32),
                                        ('voltage', np.float32),
                                        ('num_fpga', np.int32),
                                        ('num_adc', np.int32),
                                        ('pps_count', np.int32),
                                        ('dac_tune', np.int32),
                                        ('block_len', np.int32)])


    :Example:

        >>> import mtpy.usgs.zen as zen
        >>> zt = zen.Zen3D(r"/home/mt/mt00/mt00_20150522_080000_256_EX.Z3D")
        >>> zt.read_z3d()
        >>> ------- Reading /home/mt/mt00/mt00_20150522_080000_256_EX.Z3D -----
            --> Reading data took: 0.322 seconds
            Scheduled time was 2015-05-22,08:00:16 (GPS time)
            1st good stamp was 2015-05-22,08:00:18 (GPS time)
            difference of 2.00 seconds
            found 6418 GPS time stamps
            found 1642752 data points
        >>> zt.plot_time_series()
    """

    def __init__(self, fn=None, **kwargs):
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        self.fn = fn

        self.header = Z3DHeader(fn)
        self.schedule = Z3DSchedule(fn)
        self.metadata = Z3DMetadata(fn)

        self._gps_stamp_length = kwargs.pop("stamp_len", 64)
        self._gps_bytes = self._gps_stamp_length / 4

        self.gps_stamps = None

        self._gps_flag_0 = np.int32(2147483647)
        self._gps_flag_1 = np.int32(-2147483648)
        self._gps_f0 = self._gps_flag_0.tobytes()
        self._gps_f1 = self._gps_flag_1.tobytes()
        self.gps_flag = self._gps_f0 + self._gps_f1

        self._gps_dtype = np.dtype(
            [
                ("flag0", np.int32),
                ("flag1", np.int32),
                ("time", np.int32),
                ("lat", np.float64),
                ("lon", np.float64),
                ("gps_sens", np.int32),
                ("num_sat", np.int32),
                ("temperature", np.float32),
                ("voltage", np.float32),
                ("num_fpga", np.int32),
                ("num_adc", np.int32),
                ("pps_count", np.int32),
                ("dac_tune", np.int32),
                ("block_len", np.int32),
            ]
        )

        self._week_len = 604800
        # '1980, 1, 6, 0, 0, 0, -1, -1, 0
        self._gps_epoch = MTime("1980-01-06T00:00:00")
        self._leap_seconds = 18
        self._block_len = 2**16
        # the number in the cac files is for volts, we want mV
        self._counts_to_mv_conversion = 9.5367431640625e-10 * 1e3
        self.num_sec_to_skip = 2

        self.units = "counts"
        self.sample_rate = None
        self.time_series = None

        self.ch_dict = {"hx": 1, "hy": 2, "hz": 3, "ex": 4, "ey": 5}

    @property
    def fn(self):
        return self._fn

    @fn.setter
    def fn(self, fn):
        if fn is not None:
            self._fn = Path(fn)
        else:
            self._fn = None

    @property
    def file_size(self):
        if self.fn is not None:
            return self.fn.stat().st_size
        return 0

    @property
    def n_samples(self):
        if self.time_series is None:
            if self.sample_rate:
                return int(
                    (self.file_size - 512 * (1 + self.metadata.count)) / 4
                    + 8 * self.sample_rate
                )
            else:
                # assume just the 3 general metadata blocks
                return int((self.file_size - 512 * 3) / 4)
        else:
            return self.time_series.size

    @property
    def station(self):
        """
        station name
        """
        return self.metadata.station

    @station.setter
    def station(self, station):
        """
        station name
        """
        self.metadata.station = station

    @property
    def dipole_length(self):
        """
        dipole length
        """
        length = 0
        if self.metadata.ch_length is not None:
            length = float(self.metadata.ch_length)
        elif hasattr(self.metadata, "ch_offset_xyz1"):
            # only ex and ey have xyz2
            if hasattr(self.metadata, "ch_offset_xyz2"):
                x1, y1, z1 = [
                    float(offset)
                    for offset in self.metadata.ch_offset_xyz1.split(":")
                ]
                x2, y2, z2 = [
                    float(offset)
                    for offset in self.metadata.ch_offset_xyz2.split(":")
                ]
                length = np.sqrt(
                    (x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2
                )
                length = np.round(length, 2)
            else:
                length = 0
        elif self.metadata.ch_xyz1 is not None:
            x1, y1 = [float(d) for d in self.metadata.ch_xyz1.split(":")]
            x2, y2 = [float(d) for d in self.metadata.ch_xyz2.split(":")]
            length = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2) * 100.0
            length = np.round(length, 2)

        return length

    @property
    def azimuth(self):
        """
        azimuth of instrument setup
        """
        if self.metadata.ch_azimuth is not None:
            return float(self.metadata.ch_azimuth)
        elif self.metadata.rx_xazimuth is not None:
            return float(self.metadata.rx_xazimuth)
        else:
            return None

    @property
    def component(self):
        """
        channel
        """
        return self.metadata.ch_cmp.lower()

    @property
    def latitude(self):
        """
        latitude in decimal degrees
        """
        return self.header.lat

    @property
    def longitude(self):
        """
        longitude in decimal degrees
        """
        return self.header.long

    @property
    def elevation(self):
        """
        elevation in meters
        """
        return self.header.alt

    @property
    def sample_rate(self):
        """
        sampling rate
        """
        return self.header.ad_rate

    @sample_rate.setter
    def sample_rate(self, sampling_rate):
        """
        sampling rate
        """
        if sampling_rate is not None:
            self.header.ad_rate = float(sampling_rate)

    @property
    def start(self):
        if self.gps_stamps is not None:
            return self.get_UTC_date_time(
                self.header.gpsweek, self.gps_stamps["time"][0]
            )
        return self.zen_schedule

    @property
    def end(self):
        if self.gps_stamps is not None:
            return self.get_UTC_date_time(
                self.header.gpsweek, self.gps_stamps["time"][-1]
            )
        return self.start + (self.n_samples / self.sample_rate)

    @property
    def zen_schedule(self):
        """
        zen schedule data and time
        """

        if self.header.old_version is True:
            return MTime(self.header.schedule)

        return self.schedule.initial_start

    @zen_schedule.setter
    def zen_schedule(self, schedule_dt):
        """
        on setting set schedule datetime
        """
        if not isinstance(schedule_dt, MTime):
            schedule_dt = MTime(schedule_dt)
            raise TypeError(
                "New schedule datetime must be type datetime.datetime"
            )
        self.schedule.initial_start = schedule_dt

    @property
    def coil_number(self):
        """
        coil number
        """
        if self.metadata.cal_ant is not None:
            return self.metadata.cal_ant
        elif self.metadata.ch_number is not None:
            return self.metadata.ch_number
        else:
            return None

    @property
    def channel_number(self):
        if self.metadata.ch_number:
            ch_num = int(float(self.metadata.ch_number))
            if ch_num > 6:
                try:
                    ch_num = self.ch_dict[self.component]
                except KeyError:
                    ch_num = 6
            return ch_num
        else:
            try:
                return self.ch_dict[self.component]
            except KeyError:
                return 0

    @property
    def channel_metadata(self):
        """Channel metadata"""

        # fill the time series object
        if "e" in self.component:
            ch = Electric()
            ch.dipole_length = self.dipole_length
            ch.ac.start = (
                self.time_series[0 : int(self.sample_rate)].std()
                * self.header.ch_factor
            )
            ch.ac.end = (
                self.time_series[-int(self.sample_rate) :].std()
                * self.header.ch_factor
            )
            ch.dc.start = (
                self.time_series[0 : int(self.sample_rate)].mean()
                * self.header.ch_factor
            )
            ch.dc.end = (
                self.time_series[-int(self.sample_rate) :].mean()
                * self.header.ch_factor
            )
        elif "h" in self.component:
            ch = Magnetic()
            ch.sensor.id = (self.coil_number,)
            ch.sensor.manufacturer = ("Geotell",)
            ch.sensor.model = ("ANT-4",)
            ch.sensor.type = ("induction coil",)
            ch.h_field_max.start = (
                self.time_series[0 : int(self.sample_rate)].max()
                * self.header.ch_factor
            )
            ch.h_field_max.end = (
                self.time_series[-int(self.sample_rate) :].max()
                * self.header.ch_factor
            )
            ch.h_field_min.start = (
                self.time_series[0 : int(self.sample_rate)].min()
                * self.header.ch_factor
            )
            ch.h_field_min.end = (
                self.time_series[-int(self.sample_rate) :].min()
                * self.header.ch_factor
            )

        ch.time_period.start = self.start.isoformat()
        ch.time_period.end = self.end.isoformat()
        ch.component = self.component
        ch.sample_rate = self.sample_rate
        ch.measurement_azimuth = self.azimuth
        ch.units = "digital counts"
        ch.channel_number = self.channel_number
        ch.filter.name = self.channel_response.names
        ch.filter.applied = [False] * len(self.channel_response.names)

        return ch

    @property
    def station_metadata(self):
        """station metadta"""

        sm = Station()
        sm.id = self.station
        sm.fdsn.id = self.station
        sm.location.latitude = self.latitude
        sm.location.longitude = self.longitude
        sm.location.elevation = self.elevation
        sm.time_period.start = self.start.isoformat()
        sm.time_period.end = self.end.isoformat()
        sm.acquired_by.author = self.metadata.gdp_operator

        return sm

    @property
    def run_metadata(self):
        """Run metadata"""
        rm = Run()
        rm.data_logger.firmware.version = self.header.version
        rm.data_logger.id = self.header.data_logger
        rm.data_logger.manufacturer = "Zonge International"
        rm.data_logger.model = "ZEN"
        rm.time_period.start = self.start.isoformat()
        rm.time_period.end = self.end.isoformat()
        rm.sample_rate = self.sample_rate
        rm.data_type = "MTBB"
        rm.time_period.start = self.start.isoformat()
        rm.time_period.end = self.end.isoformat()
        rm.acquired_by.author = self.metadata.gdp_operator

        return rm

    @property
    def counts2mv_filter(self):
        """
        Create a counts2mv coefficient filter

        .. note:: Needs to be 1/channel factor because we divided the
         instrument response from the data.

        """

        c2mv = CoefficientFilter()
        c2mv.units_in = "digital counts"
        c2mv.units_out = "millivolts"
        c2mv.name = "zen_counts2mv"
        c2mv.gain = 1.0 / self.header.ch_factor
        c2mv.comments = "digital counts to millivolts"

        return c2mv

    @property
    def coil_response(self):
        """
        Make the coile response into a FAP filter

        Phase must be in radians
        """
        fap = None
        if self.metadata.cal_ant is not None:
            fap = FrequencyResponseTableFilter()
            fap.units_in = "millivolts"
            fap.units_out = "nanotesla"
            fap.frequencies = self.metadata.coil_cal.frequency
            fap.amplitudes = self.metadata.coil_cal.amplitude
            fap.phases = self.metadata.coil_cal.phase / 1e3
            fap.name = f"ant4_{self.coil_number}_response"
            fap.comments = "induction coil response read from z3d file"

        return fap

    @property
    def zen_response(self):
        """
        Zen response, not sure the full calibration comes directly from the
        Z3D file, so skipping for now.  Will have to read a Zen##.cal file
        to get the full calibration.  This shouldn't be a big issue cause it
        should roughly be the same for all channels and since the TF is
        computing the ratio they will cancel out.  Though we should look
        more into this if just looking at calibrate time series.

        """
        fap = None
        find = False
        if self.metadata.board_cal not in [None, []]:
            if self.metadata.board_cal[0][0] == "":
                return fap
            sr_dict = {256: 0, 1024: 1, 4096: 4}
            sr_int = sr_dict[int(self.sample_rate)]
            fap_table = self.metadata.board_cal[
                np.where(self.metadata.board_cal.rate == sr_int)
            ]
            frequency = fap_table.frequency
            amplitude = fap_table.amplitude
            phase = fap_table.phase / 1e3
            find = True

        elif self.metadata.cal_board is not None:

            try:
                fap_dict = self.metadata.cal_board[int(self.sample_rate)]
                frequency = fap_dict["frequency"]
                amplitude = fap_dict["amplitude"]
                phase = fap_dict["phase"]
                find = True
            except KeyError:
                try:
                    fap_str = self.metadata.cal_board["cal.ch"]
                    for ss in fap_str.split(";"):
                        freq, _, resp = ss.split(",")
                        ff, amp, phs = [float(item) for item in resp.split(":")]
                        if float(freq) == self.sample_rate:
                            frequency = ff
                            amplitude = amp
                            phase = phs / 1e3
                    find = True
                except KeyError:
                    return fap

        if find:
            freq = np.logspace(np.log10(6.00000e-04), np.log10(8.19200e03), 48)
            freq[np.where(freq >= frequency)[0][0]] = frequency
            amp = np.ones(48)
            phases = np.zeros(48)

            index = np.where(freq == frequency)[0][0]
            amp[index] = amplitude
            phases[index] = phase

            fap = FrequencyResponseTableFilter()
            fap.units_in = "millivolts"
            fap.units_out = "millivolts"
            fap.frequencies = freq
            fap.amplitudes = amp
            fap.phases = phases
            fap.name = f"{self.header.data_logger.lower()}_{self.sample_rate:.0f}_response"
            fap.comments = "data logger response read from z3d file"

        return None

    @property
    def channel_response(self):
        filter_list = [self.counts2mv_filter]
        if self.zen_response:
            filter_list.append(self.zen_response)
        if self.coil_response:
            filter_list.append(self.coil_response)
        elif self.dipole_filter:
            filter_list.append(self.dipole_filter)

        return ChannelResponseFilter(filters_list=filter_list)

    @property
    def dipole_filter(self):
        dipole = None
        if self.dipole_length != 0:
            dipole = CoefficientFilter()
            dipole.units_in = "millivolts"
            dipole.units_out = "millivolts per kilometer"
            dipole.name = f"dipole_{self.dipole_length:.2f}m"
            dipole.gain = self.dipole_length / 1000.0
            dipole.comments = "convert to electric field"

        return dipole

    def _get_gps_stamp_type(self, old_version=False):
        """
        get the correct stamp type.
        Older versions the stamp length was 36 bits
        New versions have a 64 bit stamp
        """

        if old_version is True:
            self._gps_dtype = np.dtype(
                [
                    ("gps", np.int32),
                    ("time", np.int32),
                    ("lat", np.float64),
                    ("lon", np.float64),
                    ("block_len", np.int32),
                    ("gps_accuracy", np.int32),
                    ("temperature", np.float32),
                ]
            )
            self._gps_stamp_length = 36
            self._gps_bytes = self._gps_stamp_length / 4
            self._gps_flag_0 = -1
            self._block_len = int(self._gps_stamp_length + self.sample_rate * 4)
            self.gps_flag = self._gps_f0

        else:
            return

    # ======================================
    def _read_header(self, fn=None, fid=None):
        """
        read header information from Z3D file
        Arguments
        ---------------
            **fn** : string
                     full path to Z3D file to read
            **fid** : file object
                      if the file is open give the file id object
        Outputs:
        ----------
            * fills the Zen3ZD.header object's attributes

        Example with just a file name
        ------------
            >>> import mtpy.usgs.zen as zen
            >>> fn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
            >>> Z3Dobj = zen.Zen3D()
            >>> Z3Dobj.read_header(fn)

        Example with file object
        ------------
            >>> import mtpy.usgs.zen as zen
            >>> fn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
            >>> Z3Dfid = open(fn, 'rb')
            >>> Z3Dobj = zen.Zen3D()
            >>> Z3Dobj.read_header(fid=Z3Dfid)
        """

        if fn is not None:
            self.fn = fn

        self.header.read_header(fn=self.fn, fid=fid)
        if self.header.old_version:
            if self.header.box_number is None:
                self.header.box_number = "6666"

    # ======================================
    def _read_schedule(self, fn=None, fid=None):
        """
        read schedule information from Z3D file
        Arguments
        ---------------
            **fn** : string
                     full path to Z3D file to read
            **fid** : file object
                      if the file is open give the file id object
        Outputs:
        ----------
            * fills the Zen3ZD.schedule object's attributes

        Example with just a file name
        ------------
            >>> import mtpy.usgs.zen as zen
            >>> fn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
            >>> Z3Dobj = zen.Zen3D()
            >>> Z3Dobj.read_schedule(fn)

        Example with file object
        ------------
            >>> import mtpy.usgs.zen as zen
            >>> fn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
            >>> Z3Dfid = open(fn, 'rb')
            >>> Z3Dobj = zen.Zen3D()
            >>> Z3Dobj.read_schedule(fid=Z3Dfid)
        """

        if fn is not None:
            self.fn = fn

        self.schedule.read_schedule(fn=self.fn, fid=fid)
        if self.header.old_version:
            self.schedule.initial_start = MTime(
                self.header.schedule, gps_time=True
            )

    # ======================================
    def _read_metadata(self, fn=None, fid=None):
        """
        read header information from Z3D file
        Arguments
        ---------------
            **fn** : string
                     full path to Z3D file to read
            **fid** : file object
                      if the file is open give the file id object
        Outputs:
        ----------
            * fills the Zen3ZD.metadata object's attributes

        Example with just a file name
        ------------
            >>> import mtpy.usgs.zen as zen
            >>> fn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
            >>> Z3Dobj = zen.Zen3D()
            >>> Z3Dobj.read_metadata(fn)

        Example with file object
        ------------
            >>> import mtpy.usgs.zen as zen
            >>> fn = r"/home/mt/mt01/mt01_20150522_080000_256_EX.Z3D"
            >>> Z3Dfid = open(fn, 'rb')
            >>> Z3Dobj = zen.Zen3D()
            >>> Z3Dobj.read_metadata(fid=Z3Dfid)
        """

        if fn is not None:
            self.fn = fn

        if self.header.old_version:
            self.metadata._schedule_metadata_len = 0
        self.metadata.read_metadata(fn=self.fn, fid=fid)

    # =====================================
    def read_all_info(self):
        """
        Read header, schedule, and metadata
        """
        with open(self.fn, "rb") as file_id:

            self._read_header(fid=file_id)
            self._read_schedule(fid=file_id)
            self._read_metadata(fid=file_id)

    def _read_raw_string(self, fid):
        """
        read raw sting into data

        :param fid: DESCRIPTION
        :type fid: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        # move the read value to where the end of the metadata is
        fid.seek(self.metadata.m_tell)

        # initalize a data array filled with zeros, everything goes into
        # this array then we parse later
        data = np.zeros(self.n_samples, dtype=np.int32)
        # go over a while loop until the data cound exceed the file size
        data_count = 0
        while True:
            # need to make sure the last block read is a multiple of 32 bit
            read_len = min(
                [
                    self._block_len,
                    int(32 * ((self.file_size - fid.tell()) // 32)),
                ]
            )
            test_str = np.frombuffer(fid.read(read_len), dtype=np.int32)
            if len(test_str) == 0:
                break
            data[data_count : data_count + len(test_str)] = test_str
            data_count += test_str.size

        return data

    def _unpack_data(self, data, gps_stamp_index):
        """ """

        for ii, gps_find in enumerate(gps_stamp_index):
            try:
                data[gps_find + 1]
            except IndexError:
                pass
                self.logger.warning(
                    f"Failed gps stamp {ii+1} out of {len(gps_stamp_index)}"
                )
                break

            if (
                self.header.old_version is True
                or data[gps_find + 1] == self._gps_flag_1
            ):
                gps_str = struct.pack(
                    "<" + "i" * int(self._gps_bytes),
                    *data[int(gps_find) : int(gps_find + self._gps_bytes)],
                )
                self.gps_stamps[ii] = np.frombuffer(
                    gps_str, dtype=self._gps_dtype
                )
                if ii > 0:
                    self.gps_stamps[ii]["block_len"] = (
                        gps_find - gps_stamp_index[ii - 1] - self._gps_bytes
                    )
                elif ii == 0:
                    self.gps_stamps[ii]["block_len"] = 0
                data[int(gps_find) : int(gps_find + self._gps_bytes)] = 0

        return data

    # ======================================
    def read_z3d(self, z3d_fn=None):
        """
        read in z3d file and populate attributes accordingly

        1. Read in the entire file as chunks as np.int32.

        2. Extract the gps stamps and convert accordingly. Check to make sure
           gps time stamps are 1 second apart and incrementing as well as
           checking the number of data points between stamps is the
           same as the sampling rate.

        3. Converts gps_stamps['time'] to seconds relative to header.gps_week
            Note we skip the first two gps stamps because there is something
            wrong with the data there due to some type of buffering.
            Therefore the first GPS time is when the time series starts, so you
            will notice that gps_stamps[0]['block_len'] = 0, this is because there
            is nothing previous to this time stamp and so the 'block_len' measures
            backwards from the corresponding time index.

        4. Put the data chunks into Pandas data frame that is indexed by time

        :Example:

        >>> from mth5.io import zen
        >>> z_obj = zen.Z3D(r"home/mt_data/zen/mt001.z3d")
        >>> z_obj.read_z3d()

        """

        if z3d_fn is not None:
            self.fn = z3d_fn

        self.logger.debug(f"Reading {self.fn}")
        st = datetime.datetime.now()

        # using the with statement works in Python versions 2.7 or higher
        # the added benefit of the with statement is that it will close the
        # file object upon reading completion.
        with open(self.fn, "rb") as file_id:

            self._read_header(fid=file_id)
            self._read_schedule(fid=file_id)
            self._read_metadata(fid=file_id)

            if self.header.old_version is True:
                self._get_gps_stamp_type(True)

            data = self._read_raw_string(file_id)

        self.raw_data = data.copy()

        # find the gps stamps
        gps_stamp_find = self.get_gps_stamp_index(data, self.header.old_version)

        # skip the first two stamps and trim data
        try:
            data = data[gps_stamp_find[self.num_sec_to_skip] :]
        except IndexError:
            msg = f"Data is too short, cannot open file {self.fn}"
            self.logger.error(msg)
            raise ZenGPSError(msg)

        # find gps stamps of the trimmed data
        gps_stamp_find = self.get_gps_stamp_index(data, self.header.old_version)

        # read data chunks and GPS stamps
        self.gps_stamps = np.zeros(len(gps_stamp_find), dtype=self._gps_dtype)
        data = self._unpack_data(data, gps_stamp_find)

        # fill the time series
        self.time_series = data[np.nonzero(data)]

        # validate everything
        self.validate_time_blocks()
        self.convert_gps_time()
        self.zen_schedule = self.check_start_time()

        self.logger.debug(f"found {self.gps_stamps.shape[0]} GPS time stamps")
        self.logger.debug(f"found {self.time_series.size} data points")

        # time it
        et = datetime.datetime.now()
        read_time = (et - st).total_seconds()
        self.logger.debug(f"Reading data took: {read_time:.3f} seconds")

    # =================================================
    def get_gps_stamp_index(self, ts_data, old_version=False):
        """
        locate the time stamps in a given time series.

        Looks for gps_flag_0 first, if the file is newer, then makes sure the
        next value is gps_flag_1

        :returns: list of gps stamps indicies
        """

        # find the gps stamps
        gps_stamp_find = np.where(ts_data == self._gps_flag_0)[0]

        if old_version is False:
            gps_stamp_find = [
                gps_find
                for gps_find in gps_stamp_find
                if ts_data[gps_find + 1] == self._gps_flag_1
            ]

        return gps_stamp_find

    # =================================================
    def trim_data(self):
        """
        apparently need to skip the first 2 seconds of data because of
        something to do with the SD buffer

        This method will be deprecated after field testing

        """
        # the block length is the number of data points before the time stamp
        # therefore the first block length is 0.  The indexing in python
        # goes to the last index - 1 so we need to put in 3
        ts_skip = self.gps_stamps["block_len"][0:3].sum()
        self.gps_stamps = self.gps_stamps[2:]
        self.gps_stamps[0]["block_len"] = 0
        self.time_series = self.time_series[ts_skip:]

    # =================================================
    def check_start_time(self):
        """
        check to make sure the scheduled start time is similar to
        the first good gps stamp
        """

        # make sure the time is in gps time
        zen_start_utc = self.get_UTC_date_time(
            self.header.gpsweek, self.gps_stamps["time"][0]
        )

        # estimate the time difference between the two
        time_diff = zen_start_utc - self.schedule.initial_start
        self.logger.debug(f"Scheduled time was {self.schedule.initial_start}")
        self.logger.debug(f"1st good stamp was {zen_start_utc}")
        self.logger.debug(f"difference of {time_diff:.2f} seconds")

        return zen_start_utc

    # ==================================================
    def validate_gps_time(self):
        """
        make sure each time stamp is 1 second apart

        """

        t_diff = np.zeros_like(self.gps_stamps["time"])

        for ii in range(len(t_diff) - 1):
            t_diff[ii] = (
                self.gps_stamps["time"][ii] - self.gps_stamps["time"][ii + 1]
            )

        bad_times = np.where(abs(t_diff) > 0.5)[0]
        if len(bad_times) > 0:
            self.logger.warning("BAD GPS TIMES:")
            for bb in bad_times:
                self.logger.warning(f"bad GPS time at index {bb} > 0.5 s")

    # ===================================================
    def validate_time_blocks(self):
        """
        validate gps time stamps and make sure each block is the proper length

        """
        # first check if the gps stamp blocks are of the correct length
        bad_blocks = np.where(
            self.gps_stamps["block_len"][1:] != self.header.ad_rate
        )[0]

        if len(bad_blocks) > 0:
            if bad_blocks.max() < 5:
                ts_skip = self.gps_stamps["block_len"][
                    0 : bad_blocks[-1] + 1
                ].sum()
                self.gps_stamps = self.gps_stamps[bad_blocks[-1] :]
                self.time_series = self.time_series[ts_skip:]

                self.logger.warning(
                    f"Skipped the first {bad_blocks[-1]} seconds"
                )
                self.logger.warning(
                    f"Skipped first {ts_skip} poins in time series"
                )

    # ==================================================
    def convert_gps_time(self):
        """
        convert gps time integer to relative seconds from gps_week

        """
        # need to convert gps_time to type float from int
        dt = self._gps_dtype.descr
        if self.header.old_version is True:
            dt[1] = ("time", np.float32)
        else:
            dt[2] = ("time", np.float32)
        self.gps_stamps = self.gps_stamps.astype(np.dtype(dt))

        # convert to seconds
        # these are seconds relative to the gps week
        time_conv = self.gps_stamps["time"].copy() / 1024.0
        time_ms = (time_conv - np.floor(time_conv)) * 1.024
        time_conv = np.floor(time_conv) + time_ms

        self.gps_stamps["time"][:] = time_conv

    # ==================================================
    def convert_counts_to_mv(self, data):
        """
        convert the time series from counts to millivolts

        """

        data *= self._counts_to_mv_conversion
        return data

    # ==================================================
    def convert_mv_to_counts(self, data):
        """
        convert millivolts to counts assuming no other scaling has been applied

        """

        data /= self._counts_to_mv_conversion
        return data

    # ==================================================
    def get_gps_time(self, gps_int, gps_week=0):
        """
        from the gps integer get the time in seconds.

        :param int gps_int: integer from the gps time stamp line
        :param int gps_week: relative gps week, if the number of seconds is
                            larger than a week then a week is subtracted from
                            the seconds and computed from gps_week += 1
        :returns: gps_time as number of seconds from the beginning of the relative
                  gps week.

        """

        gps_seconds = gps_int / 1024.0

        gps_ms = (gps_seconds - np.floor(gps_int / 1024.0)) * (1.024)

        cc = 0
        if gps_seconds > self._week_len:
            gps_week += 1
            cc = gps_week * self._week_len
            gps_seconds -= self._week_len

        gps_time = np.floor(gps_seconds) + gps_ms + cc

        return gps_time, gps_week

    # ==================================================
    def get_UTC_date_time(self, gps_week, gps_time):
        """
        get the actual date and time of measurement as UTC.


        :param int gps_week: integer value of gps_week that the data was collected
        :param int gps_time: number of seconds from beginning of gps_week

        :return: :class:`mth5.utils.mttime.MTime`

        """
        # need to check to see if the time in seconds is more than a gps week
        # if it is add 1 to the gps week and reduce the gps time by a week
        if gps_time > self._week_len:
            gps_week += 1
            gps_time -= self._week_len

        # compute seconds using weeks and gps time
        utc_seconds = (
            self._gps_epoch.epoch_seconds
            + (gps_week * self._week_len)
            + gps_time
        )

        # compute date and time from seconds and return a datetime object
        # easier to manipulate later
        return MTime(utc_seconds, gps_time=True)

    # =================================================
    def to_channelts(self):
        """
        fill time series object
        """

        return ChannelTS(
            self.channel_metadata.type,
            data=self.time_series,
            channel_metadata=self.channel_metadata,
            station_metadata=self.station_metadata,
            run_metadata=self.run_metadata,
            channel_response_filter=self.channel_response,
        )


# ==============================================================================
#  Error instances for Zen
# ==============================================================================
class ZenGPSError(Exception):
    """
    error for gps timing
    """

    pass


class ZenSamplingRateError(Exception):
    """
    error for different sampling rates
    """

    pass


class ZenInputFileError(Exception):
    """
    error for input files
    """

    pass


def read_z3d(fn, logger_file_handler=None):
    """
    generic tool to read z3d file
    """

    z3d_obj = Z3D(fn)
    if logger_file_handler:
        z3d_obj.logger.addHandler(logger_file_handler)
    try:
        z3d_obj.read_z3d()
    except ZenGPSError as error:
        z3d_obj.logger.exception(error)
        z3d_obj.logger.warning(f"Skipping {fn}, check file for GPS timing.")
        return None
    return z3d_obj.to_channelts()
